"""
Exceptions
"""
from colors import red


class InvalidDataType(Exception):
    """Exit code: 128
    This can happen when you try to pass a directory to `data` or when you
    try to pass a directory to `template`
    without '/', this can be interpreted as a file and throw this error.
    """

    def __init__(self,
                 message=f"{red('[ERROR]')} Invalid data type, try '-' for stdin, "
                 "valid URL or path on your file system."):
        self.message = message
        super().__init__(self.message)


class InvalidInput(Exception):
    """Exit code: 127"""

    def __init__(self, message=f"{red('[ERROR]')} Invalid input, "
                 "stdin can be used for data or template but not both."):
        self.message = message
        super().__init__(self.message)



"""
Data objects
"""
import json
import sys

from colors import red
from jinja2 import Template as Jinja2_template, exceptions

from .exceptions import InvalidDataType
from .plugins import plugins
from .util import print_verbose, input_handler, load_yaml
from .validators import validate_json


class Base:
    """
    Simple base class that fits it all
    """

    def __init__(self, data, verbose):
        self.data = data
        self.verbose = verbose


class ExtraVar(Base):
    """
    ExtraVar can give you the ability to specify data without
    creating any file, this can be great for debugging or overwriting
    some variables in the data file.
    """

    def __init__(self, data, verbose):
        Base.__init__(self, data, verbose)
        self.data = dict(var.split('=') for var in self.data)
        for var in self.data:
            self.data[var] = validate_json(self.data[var])

    def get_extra_vars(self):
        """
        A simple getter with verbose option
        """
        print_verbose({'title': '[ExtraVars]', 'content': json.dumps(
            self.data, indent=2), 'verbose': self.verbose})
        return self.data


class Data(Base):
    """
    A class that holds the data
    """

    def __init__(self, data, verbose):
        Base.__init__(self, data, verbose)

    def get_data(self):
        """
        Decide how to interpret the data & load it in a way that the
        Template class can render.
        For now it can load both YAMLs & JSONs.
        """
        try:
            ih_content = input_handler(self.data)
        except InvalidDataType as err:
            print(err.message, file=sys.stderr)
            sys.exit(128)
        else:
            self.data = load_yaml(ih_content)
            print_verbose({'title': '[Data]', 'content': json.dumps(
                self.data, indent=2), 'verbose': self.verbose})
            return self.data


class Template(Base):
    """
    Used to render the template using the data
    """

    def __init__(self, template, verbose, data, options):
        Base.__init__(self, data, verbose)
        self.template = template
        self.no_trim_blocks = options['no_trim_blocks']
        self.no_lstrip_blocks = options['no_lstrip_blocks']

    def __render(self):
        """
        Decide whether to render with data or not.
        There is an option to render using only extra variables.
        """
        try:
            self.data.update(plugins())
            return self.template.render(self.data)
        except exceptions.UndefinedError as err:
            print(red('[ERROR]'), err.message, file=sys.stderr)

    def get_rendered_template(self):
        """
        Decide how to interpret the template, and return its rendered version.
        """
        try:
            ih_content = input_handler(self.template)
        except InvalidDataType as err:
            print(err.message, file=sys.stderr)
            sys.exit(128)
        else:
            print_verbose(
                {'title': '[Template]', 'content': ih_content, 'verbose': self.verbose})
            self.template = Jinja2_template(
                ih_content,
                trim_blocks=self.no_trim_blocks,
                lstrip_blocks=self.no_lstrip_blocks,
            )
            return self.__render()



"""
Utility
"""
import os
import sys
import requests
import yaml

from colors import green

from .validators import validate_url, validate_is_file
from .exceptions import InvalidDataType


def print_verbose(message):
    """Print verbose output to stdout"""
    if message['verbose']:
        separator = green(f'{"-" * 10}')
        print(separator, message['title'], separator)
        print(green(message['content']), "\n")


def load_yaml(data):
    """Load YAML / JSON"""
    return yaml.load(data, Loader=yaml.FullLoader)


def output_template(content, output_path, relative_path=None):
    """Print template to stdout by default
    If output_path is not None, write the template content
    to the path that was specified.
    If relative_path is not None, copy the template directory with all
    templates applied.
    """
    if content is not None:
        if output_path is None:
            print(content)
        else:
            if relative_path is not None:
                relative_path = output_path + relative_path
                if relative_path is not os.path.exists(os.path.dirname(relative_path)):
                    os.makedirs(os.path.dirname(relative_path), exist_ok=True)
                output_path = relative_path
            with open(output_path, 'w+') as output_file:
                output_file.write(content)


def input_handler(data):
    """Handle & validate input"""
    ih_content = None
    if data == '-':
        ih_content = sys.stdin.read()
    elif validate_url(data):
        response = requests.get(data)
        if not response.raise_for_status():
            ih_content = response.text
    elif validate_is_file(data):
        with open(data, 'r') as input_data_file:
            ih_content = input_data_file.read()
    else:
        raise InvalidDataType()
    return ih_content



"""
Validators
"""
import json
import os
import validators


def validate_json(j):
    """Try to parse as JSON, if it fails return it as a raw string"""
    try:
        ans = json.loads(j)
    except json.decoder.JSONDecodeError:
        return j
    else:
        return ans


def validate_url(url) -> bool:
    """Try to parse as URL"""
    return validators.url(url)


def validate_path(path) -> bool:
    """Validate that the path exist & have permisions to read from"""
    return os.path.exists(path) and \
        os.access(os.path.dirname(path), os.R_OK)


def validate_is_file(path) -> bool:
    """validate_path(path) and is a file"""
    return validate_path(path) and os.path.isfile(path)


def validate_is_dir(path) -> bool:
    """validate_path(path) and is a directory"""
    return validate_path(path) and os.path.isdir(path)



import os
import sys
import requests
import shutil
import yaml

from filecmp import dircmp
from io import StringIO

from click.testing import CliRunner
from jinja2_tools.cli import main as jinja

TEMPLATE_URL = 'https://raw.githubusercontent.com/Avielyo10/jinja2-tools/master/examples/templates/template.yaml'
DATA_URL = 'https://raw.githubusercontent.com/Avielyo10/jinja2-tools/master/examples/data/data.yaml'
EXTRA_VARS = 'access_lists={\"al-hq-in\": [{\"action\": \"remark\", \"text\": \"Allow traffic from hq to local office\"}, {\"action\": \"permit\", \"src\": \"10.0.0.0/22\", \"dst\": \"10.100.0.0/24\"}]}'

DEFAULT_MSG = """(1)
ip access-list extended al-hq-in
(2)
    (3)    remark Allow traffic from hq to local office
    (4)(2)
    (3)    permit 10.0.0.0/22 10.100.0.0/24
    (5)(6)
(7)
hello jinja!
# All ACLs have been generated
"""

MSG_WITH_LB_TB = """(1)
ip access-list extended al-hq-in
  (2)
    (3)
    remark Allow traffic from hq to local office
    (4)
  (2)
    (3)
    permit 10.0.0.0/22 10.100.0.0/24
    (5)
  (6)
(7)
hello jinja!
# All ACLs have been generated
"""

DIR_COMPARISONS="""diff test/ examples/
Common subdirectories : ['config', 'data', 'templates']

diff test/config examples/config
Identical files : ['example.ini', 'example.properties']

diff test/data examples/data
Identical files : ['data.json', 'data.yaml']

diff test/templates examples/templates
Differing files : ['lookup.yaml', 'template.yaml']
"""
os.environ['SHELL'] = '/usr/bin/zsh'


def test_simple_render():
    runner = CliRunner()
    yaml_result = runner.invoke(
        jinja, ['render', '-d', 'examples/data/data.yaml', '-t', 'examples/templates/template.yaml'])
    assert yaml_result.exit_code == 0
    assert yaml_result.output == DEFAULT_MSG

    json_result = runner.invoke(
        jinja, ['render', '-d', 'examples/data/data.json', '-t', 'examples/templates/template.yaml'])
    assert json_result.output == yaml_result.output


def test_simple_render_with_tb_lb():
    runner = CliRunner()
    yaml_result = runner.invoke(
        jinja, ['render', '-lb', '-tb', '-d', 'examples/data/data.yaml', '-t', 'examples/templates/template.yaml'])
    assert yaml_result.exit_code == 0
    assert yaml_result.output == MSG_WITH_LB_TB

    json_result = runner.invoke(
        jinja, ['render', '-lb', '-tb', '-d', 'examples/data/data.json', '-t', 'examples/templates/template.yaml'])
    assert json_result.output == yaml_result.output


def test_render_with_url():
    runner = CliRunner()
    result = runner.invoke(
        jinja, ['render', '-d', DATA_URL, '-t', TEMPLATE_URL])
    assert result.exit_code == 0
    assert result.output == DEFAULT_MSG


def test_render_with_stdin_data():
    runner = CliRunner()
    result = runner.invoke(
        jinja, ['render', '-d', '-', '-t', TEMPLATE_URL], input=requests.get(DATA_URL).text)
    assert result.exit_code == 0
    assert result.output == DEFAULT_MSG


def test_render_with_stdin_template():
    runner = CliRunner()
    result = runner.invoke(
        jinja, ['render', '-d', DATA_URL, '-t', '-'], input=requests.get(TEMPLATE_URL).text)
    assert result.exit_code == 0
    assert result.output == DEFAULT_MSG


def test_render_with_extra_vars():
    runner = CliRunner()
    result = runner.invoke(
        jinja, ['render', '-d', 'examples/data/data.yaml', '-t', 'examples/templates/template.yaml', '-e', EXTRA_VARS])

    assert result.exit_code == 0
    assert result.output == DEFAULT_MSG


def test_render_with_extra_vars_override():
    runner = CliRunner()
    result = runner.invoke(
        jinja, ['render', '-d', 'examples/data/data.yaml', '-t', 'examples/templates/template.yaml', '-e', EXTRA_VARS, '-e', 'message=world'])

    assert result.exit_code == 0
    assert 'hello world!' in result.output


def test_render_with_directory():
    runner = CliRunner()
    test_dir = 'test/'
    result = runner.invoke(
        jinja, ['render', '-d', 'examples/data/data.yaml', '-t', 'examples/', '-o', test_dir])

    assert result.exit_code == 0
    ### report_full_closure() just print to stdout, so we capture it
    s_io = StringIO()
    sys.stdout = s_io
    dircmp(test_dir, 'examples/').report_full_closure()
    sys.stdout = sys.__stdout__
    s_io.seek(0, os.SEEK_SET)
    assert s_io.read() == DIR_COMPARISONS
    try:
        shutil.rmtree(test_dir)
    except FileNotFoundError as err:
        print(err.strerror)


def test_invalid_input():
    runner = CliRunner()
    yaml_result = runner.invoke(
        jinja, ['render', '-d', '-', '-t', '-'])
    assert yaml_result.exit_code == 127


def test_invalid_data_type():
    runner = CliRunner()
    test_dir = 'test/'
    result = runner.invoke(
        jinja, ['render', '-d', 'examples/', '-t', 'examples/', '-o', test_dir])

    assert result.exit_code == 128
    try:
        shutil.rmtree(test_dir)
    except FileNotFoundError:
        pass


def test_render_with_bad_url():
    runner = CliRunner()
    result = runner.invoke(
        jinja, ['render', '-d', DATA_URL+'/bla', '-t', TEMPLATE_URL])
    assert result.exit_code != 0

def test_lookup():
    runner = CliRunner()
    test_dir = 'test/'
    result = runner.invoke(
        jinja, ['render', '-d', 'examples/data/data.yaml', '-t', 'examples/', '-o', test_dir])

    assert result.exit_code == 0

    with open('test/templates/lookup.yaml', 'r') as f:
        out = yaml.full_load(f)
    lookup = out['lookup']
    
    assert lookup[0] == '/usr/bin/zsh'
    assert lookup[1] == 'gertrude'
    assert lookup[2] == ['john', 'passwd']
    assert lookup[3] == 'robert'
    assert lookup[4] == ['robert', 'somerandompassword']
    try:
        shutil.rmtree(test_dir)
    except FileNotFoundError as err:
        print(err.strerror)


"""
Lookup plugin
"""
from io import StringIO
import os
import sys
import re
import configparser

from colors import red

from jinja2_tools.validators import validate_is_file


class Lookup:
    """
    Base class of lookup plugin
    """

    def __init__(self, args):
        if len(args) == 2:
            self.op, self.arg = args
        else:
            raise ValueError(args)


class LookupResolver:
    def resolve(_, lookup):
        """
        Get the right resolver and act accordingly
        """
        resolver = get_resolver(lookup.op)
        return resolver(lookup)


def get_resolver(op):
    """
    Get the right resolver
    """
    if op == 'env':
        return _resolve_env
    elif op == 'file':
        return _resolve_file
    elif op == 'ini':
        return _resolve_ini
    else:
        raise ValueError(op)


def handle_ini_args(ini_args):
    try:
        key = ini_args[0]  # The first value in the argument is the key
        ini_args = ini_args[1:]
    except ValueError:
        raise ValueError(ini_args)
    else:
        # Default args
        args = {
            'type': 'ini',
            'file': 'jinja.ini',
            'section': 'global',
            're': False,
            'default': ''
        }
        args['key'] = key
        args.update(dict(arg.split('=') for arg in ini_args))
        return args


def _resolve_ini(lookup):
    ini_args = handle_ini_args(lookup.arg.split(' '))
    config_parser = configparser.ConfigParser()
    file = ini_args['file']
    if not validate_is_file(file):
        raise ValueError(file)

    config = StringIO()
    if ini_args['type'] == 'properties':
        config.write('[java_properties]\n')
        ini_args['section'] = 'java_properties'

    with open(file, 'r') as ini_file:
        config.write(ini_file.read())
        config.seek(0, os.SEEK_SET)
    
    config_parser.read_file(config)
    if ini_args['re']:
        return [value for key, value in config_parser.items(ini_args['section']) if re.match(ini_args['key'], key)]
    return config_parser.get(ini_args['section'], ini_args['key'], fallback=ini_args['default'])


def _resolve_file(lookup):
    """
    Implement file resolver
    """
    current_pwd = os.getcwd()
    if validate_is_file(os.path.join(current_pwd, lookup.arg)):
        with open(lookup.arg, 'r') as in_file:
            return in_file.read()
    else:
        raise ValueError(f'{lookup.arg} is not a file.')


def _resolve_env(lookup):
    """
    Implement environment resolver
    """
    try:
        key = os.environ[lookup.arg]
    except OSError as err:
        print(red('[ERROR]'), err.strerror, file=sys.stderr)
    else:
        return key



from .lookup import Lookup, LookupResolver


def lookup(*args):
    lookup_resolver = LookupResolver()
    lookup = Lookup(args)
    return lookup_resolver.resolve(lookup)


def plugins():
    plugins = {
        'lookup': lookup
    }
    return plugins



"""
setup.py for using pip
"""
import setuptools

with open("README.md", "r", encoding="utf-8") as readme_file:
    long_description = readme_file.read()

setuptools.setup(
    name="jinja2-tools",
    version="1.0.6",
    author="Aviel Yosef",
    author_email="Avielyo10@gmail.com",
    description="Use Jinja2 templates via cli",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Avielyo10/jinja2-tools.git",
    packages=setuptools.find_packages(),
    install_requires=[
        'Click',
        'ansicolors',
        'PyYAML',
        'Jinja2',
        'requests',
        'validators'
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.5',
    entry_points='''
        [console_scripts]
        jinja=jinja2_tools.cli:main
    ''',
)



"""
jinja cli
"""

import os
import sys
import traceback
import click

from jinja2 import Environment, FileSystemLoader

from .exceptions import InvalidInput
from .objects import Data, Template, ExtraVar
from .util import output_template
from .validators import validate_is_dir


@click.group()
@click.version_option()
def main():
    """
    Main click group for future commands
    """
    pass


@main.command()
@click.option('--data', '-d', help="PATH to YAML or JSON data file, URL or '-' for stdin.")
@click.option('--template', '-t', help="PATH to directory or to"
              " any file that uses Jinja, URL or '-' for stdin.")
@click.option('--verbose', '-v', is_flag=True)
@click.option('--no-trim-blocks', '-tb', default=True, is_flag=True, help='Disable trim blocks.')
@click.option('--no-lstrip-blocks', '-lb', default=True, is_flag=True,
              help='Disable lstrip blocks.')
@click.option('--output', '-o', type=click.Path(), help='PATH for output, stdout by default.')
@click.option('--extra-var', '-e', multiple=True, help="key value pair separated by '='. "
              "'value' will be treated as JSON or as a string in case of JSON decoding error. "
              "This will take precedence over 'data'.")
def render(data, template, verbose, no_trim_blocks, no_lstrip_blocks, output, extra_var):
    """
    Render templates using Jinja2
    You can render a template (or a directory) from URL, PATH or stdin,
    using data from a URL, file or stdin.
    """
    if data == '-' and template == '-':
        try:
            raise InvalidInput()
        except:
            traceback.print_exc()
            sys.exit(127)

    if data is not None:
        data = Data(data, verbose).get_data()
    else:
        data = dict()

    if extra_var is not None:
        extra_var = ExtraVar(extra_var, verbose).get_extra_vars()
        data.update(extra_var)

    if template is not None:
        options = {'no_trim_blocks': no_trim_blocks,
                   'no_lstrip_blocks': no_lstrip_blocks}

        if validate_is_dir(template):
            env = Environment(loader=FileSystemLoader(template))
            for path in env.list_templates():
                template_path = os.path.join(template, path)
                out = Template(template_path, verbose, data,
                               options).get_rendered_template()
                output_template(
                    content=out, output_path=output,
                    relative_path=path)
        else:
            out = Template(template, verbose, data,
                           options).get_rendered_template()
            output_template(out, output)
